# Copyright (c) 2023 Arista Networks, Inc.
# Use of this source code is governed by the Apache License 2.0
# that can be found in the LICENSE file.
from __future__ import annotations

import re
from collections import ChainMap
from functools import cached_property

from pyavd.vendor.j2.filter.range_expand import range_expand
from pyavd.vendor.errors import AristaAvdError
from pyavd.vendor.strip_empties import strip_null_from_data
from pyavd.vendor.utils import append_if_not_duplicate, default, get, replace_or_append_item

from .utils import UtilsMixin


class EthernetInterfacesMixin(UtilsMixin):
    """
    Mixin Class used to generate structured config for one key.
    Class should only be used as Mixin to a AvdStructuredConfig class
    """

    @cached_property
    def ethernet_interfaces(self) -> list | None:
        """
        Return structured config for ethernet_interfaces

        Duplicate checks following these rules:
        - Silently overwrite duplicate network_ports with other network_ports.
        - Silently overwrite duplicate network_ports with connected_endpoints.
        - Do NOT overwrite connected_endpoints with other connected_endpoints. Instead we raise a duplicate error.
        """

        ethernet_interfaces = []

        # List of ethernet_interfaces used for duplicate checks.

        non_overwritable_ethernet_interfaces = []

        for network_port in self._filtered_network_ports:
            connected_endpoint = {
                "name": network_port.get("description"),
                "type": "network_port",
            }
            for ethernet_interface_name in range_expand(network_port["switch_ports"]):
                # Override switches and switch_ports to only render for a single interface
                tmp_network_port = ChainMap(
                    {
                        "switch_ports": [ethernet_interface_name],
                        "switches": [self.shared_utils.hostname],
                    },
                    network_port,
                )
                ethernet_interface = self._get_ethernet_interface_cfg(tmp_network_port, 0, connected_endpoint)
                replace_or_append_item(ethernet_interfaces, "name", ethernet_interface)

        for connected_endpoint in self._filtered_connected_endpoints:
            for adapter in connected_endpoint["adapters"]:
                for node_index, node_name in enumerate(adapter["switches"]):
                    if node_name != self.shared_utils.hostname:
                        continue

                    ethernet_interface = self._get_ethernet_interface_cfg(adapter, node_index, connected_endpoint)
                    append_if_not_duplicate(
                        list_of_dicts=non_overwritable_ethernet_interfaces,
                        primary_key="name",
                        new_dict=ethernet_interface,
                        context="Ethernet Interfaces defined under connected_endpoints",
                        context_keys=["name", "peer_interface"],
                    )

                    replace_or_append_item(ethernet_interfaces, "name", ethernet_interface)

        if ethernet_interfaces:
            return ethernet_interfaces

        return None

    def _get_ethernet_interface_cfg(self, adapter: dict, node_index: int, connected_endpoint: dict) -> dict:
        """
        Return structured_config for one ethernet_interface
        """
        peer = connected_endpoint["name"]
        endpoint_ports: list = default(
            adapter.get("endpoint_ports"),
            [],
        )
        peer_interface = endpoint_ports[node_index] if node_index < len(endpoint_ports) else None
        default_channel_group_id = int("".join(re.findall(r"\d", adapter["switch_ports"][0])))
        channel_group_id = get(adapter, "port_channel.channel_id", default=default_channel_group_id)
        short_esi = self._get_short_esi(adapter, channel_group_id)

        # check lengths of lists
        nodes_length = len(adapter["switches"])
        if len(adapter["switch_ports"]) != nodes_length or ("descriptions" in adapter and len(adapter["descriptions"]) != nodes_length):
            raise AristaAvdError(
                f"Length of lists 'switches', 'switch_ports', and 'descriptions' (if used) must match for adapter. Check configuration for {peer}, adapter"
                f" switch_ports {adapter['switch_ports']}."
            )

        # if 'descriptions' is set, it is preferred
        if (interface_descriptions := adapter.get("descriptions")) is not None:
            interface_description = interface_descriptions[node_index]
        else:
            interface_description = adapter.get("description")

        # Common ethernet_interface settings
        ethernet_interface = {
            "name": adapter["switch_ports"][node_index],
            "peer": peer,
            "peer_interface": peer_interface,
            "peer_type": connected_endpoint["type"],
            "port_profile": adapter.get("profile"),
            "description": self.shared_utils.interface_descriptions.connected_endpoints_ethernet_interfaces(peer, peer_interface, interface_description),
            "speed": adapter.get("speed"),
            "shutdown": not adapter.get("enabled", True),
            "eos_cli": adapter.get("raw_eos_cli"),
            "struct_cfg": adapter.get("structured_config"),
        }

        # Port-channel member
        if (port_channel_mode := get(adapter, "port_channel.mode")) is not None:
            ethernet_interface.update(
                {
                    "type": "port-channel-member",
                    "channel_group": {
                        "id": channel_group_id,
                        "mode": port_channel_mode,
                    },
                }
            )
            if get(adapter, "port_channel.lacp_fallback.mode") == "static":
                ethernet_interface["lacp_port_priority"] = 8192 if node_index == 0 else 32768
            if port_channel_mode != "on" and get(adapter, "port_channel.lacp_timer") is not None:
                ethernet_interface["lacp_timer"] = {
                    "mode": get(adapter, "port_channel.lacp_timer.mode"),
                    "multiplier": get(adapter, "port_channel.lacp_timer.multiplier"),
                }

        # NOT a port-channel member
        else:
            ethernet_interface.update(
                {
                    "type": "switched",
                    "mtu": adapter.get("mtu") if self.shared_utils.platform_settings_feature_support_per_interface_mtu else None,
                    "l2_mtu": adapter.get("l2_mtu"),
                    "mode": adapter.get("mode"),
                    "vlans": adapter.get("vlans"),
                    "trunk_groups": self._get_adapter_trunk_groups(adapter, connected_endpoint),
                    "native_vlan_tag": adapter.get("native_vlan_tag"),
                    "native_vlan": adapter.get("native_vlan"),
                    "spanning_tree_portfast": adapter.get("spanning_tree_portfast"),
                    "spanning_tree_bpdufilter": adapter.get("spanning_tree_bpdufilter"),
                    "spanning_tree_bpduguard": adapter.get("spanning_tree_bpduguard"),
                    "storm_control": self._get_adapter_storm_control(adapter),
                    "service_profile": adapter.get("qos_profile"),
                    "dot1x": adapter.get("dot1x"),
                    "poe": self._get_adapter_poe(adapter),
                    "ptp": self._get_adapter_ptp(adapter),
                    "sflow": self._get_adapter_sflow(adapter),
                    "evpn_ethernet_segment": self._get_adapter_evpn_ethernet_segment_cfg(
                        adapter, short_esi, node_index, connected_endpoint, "auto", "single-active"
                    ),
                    "link_tracking_groups": self._get_adapter_link_tracking_groups(adapter),
                }
            )

        # More common ethernet_interface settings
        if (flowcontrol_received := get(adapter, "flowcontrol.received")) is not None:
            ethernet_interface["flowcontrol"] = {"received": flowcontrol_received}

        return strip_null_from_data(ethernet_interface)
